#!/usr/local/bin/ruby
# $Id: eparser.rb,v 1.24 2002/11/21 13:45:14 abe Exp $
#                                                      Project Gemini
#---------------------------------------------------------------------
#
#   Earley Parser
#
#      Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp>
######################################################################
require File::join( File.dirname(__FILE__), "nlputil.rb" )


#
#   Used Exception Objects
######################################################################

class TooManyDataException < NLPException
  # f[^ꍇɔO
end

class NoGrammarGivenException < RuntimeError
  # \͂ɕKvȕ@^ĂȂꍇɔO
end


#
#   Core Object
######################################################################

class EarleyParserUtility < NLPModule

  DefaultStartName = '<start>'

  VERSION = <<"--------"
#{GEMINI::mklogo("Earley Parser   ver. 2.0")}
        Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp>

--------

  USAGE = <<"--------"
usage:
  ruby eparser.rb (options) (corpus files)

options:
  --grammar FILE              using grammar
  -g FILE

  --start NAME                start symbol
  -s NAME                       (default = #{DefaultStartName})

  --simply                    output <dItemSet> only

  --id STRING                 specify id attribute of <pCorpus>

  --inspect                   output subtrees when failed
      --inspect-depth SIZE         depth of output subtree
      --inspect-all                output all patterns
      --inspect-wordnum            <A>(0,1) --> <A>[1]
--------

  # RXgN^
  #----------------------------------------
  def initialize ()
    super()
    @VERSION = VERSION
    @USAGE = USAGE
  end
  
  # IvVf[^CX^Xϐ
  #----------------------------------------
  def scan_argument ( argtbl )
    @GrammarFilename = String(argtbl['grammar'])
    
    @Simply = argtbl['simply'] || false
    
    @Inspect = argtbl['inspect'] || false
    
    @InspectDepth = argtbl['inspect-depth'].to_i # @InspectDepth == 0 ͖
    
    @InspectAll = argtbl['inspect-all'] || false
    
    @InspectWordnum = argtbl['inspect-wordnum'] || false
    
    @Id = String(argtbl['id'])
    
    @StartName = argtbl['start'] || DefaultStartName
    @StartName = "<#{@StartName}>" unless /\A<.*>\z/ =~ @StartName
  end
  
  # [eBeBIuWFNg̎s
  #----------------------------------------
  def call_body ( argtbl )
    scan_argument( argtbl )
    @ResultLog = LogBlock.new()
    
    @total = @success = @failure = 0
    @totaltime = Struct::Tms.new( 0, 0, 0, 0 )
    
    @grammar = read_grammar(@GrammarFilename)
    raise(NoGrammarGivenException, "no grammar given") if !@grammar
    
    @Start = @grammar.getcat(@StartName)
    raise(UnknownSymbolException, "unknown symbol ``#{@StartName}''") unless @Start
    v_print(" + start symbol = #{@Start} (\##{@Start.id})\n")
    
    parser = Ext::EarleyParser.new(@grammar)
    parser.start = @Start
    
    attr_override = (@Id.empty?)? nil: {'id'=>@Id}
    process_multi_corpus(@inputs, attr_override) {|corpus| process_corpus(parser, corpus) }
    
    @ResultLog.update(
      'parse-time' => "#{@totaltime.ptime} \# sec",
      'corpus-size' => @total,
      'success-count' => @success,
      'failure-count' => @failure
    )
    @ResultLog.display(@output)
    v_print(" + OK.\n")
    
    return self
  end
  def process_corpus ( parser, corpus )
    print_progress = lambda { v_print( sprintf("   - \#%d (%.1f sec)\r", @total, @totaltime.ptime) ) }
    
    t = 0
    corpus.each do |sentence|
      @total += 1
      t = update_progress_info( t ) { print_progress.call() }
      sentence.display(@output) do
        begin
          # ̓ǂݍ
          wordtree, logblock, textlist = read_sentence( sentence )
          break unless wordtree
          
          # \
          wordtree.chart = chart = time_block { parser.parse( wordtree ) }
          @totaltime += $time
          is_success = chart.exist?
          
          # \͂ɐĂ dItemSet o
          if is_success then
            @success += 1
            if !@Simply then
              # LZꂽڐȂǂ͐mɌvȂ̂ŏo͒~
              #logblock.update({ ... }).display(@output)
            end
            ItemSetBlock::display_itemset( @Start, chart, nil, @output )
          end
          
          # Kvɉĕ؂o
          if @Inspect then
            Block::display('dSubItemSet', nil, @output) do
              chartlen = chart.length - 1
              if textlist and chartlen != textlist.length then
                print_warn(corpus.stream, "\##{@total}: sentence length = #{chartlen}, text length=#{textlist.length}")
                textlist = nil
              end
              inspect_chart( chart, textlist, wordtree.terminals )
            end
          end
          
          # \͂ɎsĂΗO𓊂
          if !is_success then
            raise(RuntimeError, "parsing failed")
          end
        rescue
          @failure += 1
          print_warn(corpus.stream, "\##{@total}: #$!")
          Block::display('dLog', nil, @output) do
            @output << "ERROR : #$!\n"
          end
          t = 0
        end
      end # sentence.display
    end # corpus.each
    print_progress.call() ; v_print("\n")
  end
  attr_reader :total      # ǂݍ񂾕̑
  attr_reader :totaltime  # vԂ̑v
  attr_reader :success    # \͂̐
  attr_reader :failure    # \͂̎s
  
  # <pSentence> ̓ǂݍŕKvȏW߂
  #----------------------------------------
  def read_sentence ( sentence )
    wordtree = nil
    textlist = nil
    logblock = LogBlock.new()
    
    out = (@Simply)? nil: @output
    sentence.out = out
    sentence.each( PoSBlock, LogBlock, TextBlock ) do |cls, block|
      if cls then
        if cls == PoSBlock then
          raise(NLPException, "too many <dPoS> block") if wordtree
          wordtree = block.display(out) { cls.new(block).read(@grammar) }
        elsif cls == LogBlock then
          block.out = nil # ŏo͂邽߃O͏o͂Ȃ
          logblock.update( cls.new(block).table )
        elsif cls == TextBlock then
          textlist = block.display(out) { cls.new(block).read }
        end
      else
        if ItemSetBlock === block then
          block.skip()
        else
          block.dump(out)
        end
      end
    end
    return [wordtree, logblock, textlist]
  end
  
  # \؏o͂
  #----------------------------------------
  def inspect_chart ( chart, textlist, terminals )
    @textlist = Array(textlist) # LXgp
    @terminals = Array(terminals) # FIRST Wɂ锻ʗp
    @spacer = [ ]
    
    subtrees = chart.list_subtrees
    print_subtree(subtrees.shift, 0) until subtrees.empty?
    return self
  end
  def print_subtree ( subtree, level )
    children = subtree.children
    return unless children
    
    if children.length > 1 and @InspectAll then
      # ʂ肠ꍇ ( --inspect-all w肳Ăꍇ)
      swap_spacer_tail( "+-" ) { print_label_line( subtree.top, subtree.left, subtree.right ) }
      
      @spacer.push("    ")
      children.each_with_index do |pattern, pattern_no|
        @spacer.push( (pattern_no < children.length - 1)? ":   ": "    " )
        print_subtree_children( subtree, pattern, level, "==> " )
        @spacer.pop()
      end
      @spacer.pop()
    else
      # ʂ肵Ȃꍇ ( --inspect-all w肳ĂȂꍇ)
      if children.length > 1 then
        print_subtree_children( subtree, children[ rand(children.length) ], level, " *- " )
      else
        print_subtree_children( subtree, children.first, level, " +- " )
      end
    end
  end
  def print_subtree_children ( element, children, level, header )
    # [ĂȂꍇɕsvȍo͂Kv͂Ȃ
    unless element.complete? then
      if @grammar.has_FIRST? and element.right < @terminals.length then
        # FIRST WŔʂłĂ܂ꍇ
        return if Array(@terminals[element.right]).find {|e| @grammar.reachable?( element.top, e ) }
      end
    end
    
    # ؃gbṽx͒ŐL header ɕύXĕ\
    swap_spacer_tail( header ) { print_label_line( element.top, element.left, element.right ) }
    
    return if !children or (@InspectDepth > 0 and level >= @InspectDepth)
    
    rhs = element.rule.rhs
    @spacer.push( nil )
    for i in 0...rhs.length do
      if i >= children.length then
        # [ĂȂL
        @spacer[-1] = " :... "
        print_space(); print "#{rhs[i]}\n"
      else
        # [ĂL
        @spacer[-1] = (i < rhs.length - 1)? " |  ": "    "
        print_subtree( children[i], level + 1 )
      end
    end
    @spacer.pop
    print "\n" if level == 0
  end
  def swap_spacer_tail ( string )
    # ꎞI @spacer ̖w肳ꂽɓꊷăubNs
    if @spacer.empty? then
      yield()
    else
      tmp = @spacer[-1]
      @spacer[-1] = string
      yield()
      @spacer[-1] = tmp
    end
  end
  def print_space ()
    @spacer.each {|s| print s }
  end
  def get_labelstr ( top, left, right )
    # x𐶐
    terms = Array(@terminals[left]).join(',')
    
    string = 
      if @InspectWordnum then
        # Pꎩ̂Ɉʒuuꍇ
        if left + 1 == right then
          "#{top} [#{right}]  |#{terms}|"
        else
          "#{top} [#{left+1}..#{right}]"
        end
      else
        # PԂɈʒuuꍇ
        if right - left == 1 then
          "#{top} (#{left}, #{right})  |#{terms}|"
        else
          "#{top} (#{left}, #{right})"
        end
      end
    
    unless @textlist.empty? then
      string << "  {#{@textlist[left...right].join(' ')}}"
    end
    
    return string
  end
  def print_label_line ( top, left, right )
    # xso͂
    print_space()
    print get_labelstr(top, left, right)
    print "\n"
  end

end



######################################################################
if $0 == __FILE__ then
### ̃t@Ĉ܂܎sꂽꍇ̏
print_error_message = lambda {|mes| $stderr << "ERROR: #{mes}\n" }

begin
  optspec = [ ]
  optspec << ['--start', '-s', ARG_RQ]
  optspec << ['--grammar', '-g', ARG_RQ]
  optspec << ['--simply', ARG_NO]
  optspec << ['--id', ARG_RQ]
  optspec << ['--inspect', ARG_NO]
  optspec << ['--inspect-depth', ARG_RQ]
  optspec << ['--inspect-all', ARG_NO]
  optspec << ['--inspect-wordnum', ARG_NO]
  
  core = EarleyParserUtility.new()
  argtbl = { }
  inputs = core.fetch_options( ARGV, argtbl, optspec )
  if inputs then
    begin
      core.call( argtbl, inputs )
    rescue NoGrammarGivenException
      print_error_message.call( $! )
      core.print_usage( $stdout )
    end
  end
rescue
  print_error_message.call( $! )
end

###
end
######################################################################
