#!/usr/local/bin/ruby
# $Id: sparser.rb,v 1.2 2002/11/22 02:59:44 abe Exp $
#                                                      Project Gemini
#---------------------------------------------------------------------
#
#   Stolcke Parser
#
#      Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp>
######################################################################
require File::join( File.dirname(__FILE__), "nlputil.rb" )


#
#   Used Exception Objects
######################################################################

class TooManyDataException < RuntimeError
  # f[^ꍇɔO
end

class NoDataGivenException < RuntimeError
  # \͂ɕKvȃf[^^ĂȂꍇɔO
end

class InvalidAlgorithmException < RuntimeError
  # ȃASY^ꂽƂɔO
end


#
#   Core Object
######################################################################

class StolckeParserUtility < NLPModule
  
  DefaultStartName = '<start>'
  DefaultAlgoName = 'fixed'
  DefaultThreshold = 0.000001
  DefaultLoopcount = 0
  
  # EM wK̃Of[^
  EMHistory = Struct.new(:loopcount, :likelihood, :time, :sec)
  
  VERSION = <<"--------"
#{GEMINI::mklogo("Stolcke Parser   ver. 1.0")}
        Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp>

--------

  USAGE = <<"--------"
usage:
  ruby sparser.rb (options) (corpus files)

options:
  --grammar FILE              using grammar
  -g FILE

  --param FILE                parameter file
  -p FILE

  --algo NAME                 using algorithm
  -a NAME                       (default = #{DefaultAlgoName})
    algorithm names:
      fixed, flq, original

  --epsilon THRESHOLD         EM loop threshold
  -e THRESHOLD                  (default = #{DefaultThreshold})

  --count NUM                 EM loop count
  -c NUM                        (default = #{DefaultLoopcount})

--------
  
  # RXgN^
  #----------------------------------------
  def initialize ()
    super()
    @VERSION = VERSION
    @USAGE = USAGE
  end
  
  # IvVf[^CX^Xϐ
  #----------------------------------------
  def scan_argument ( argtbl )
    @GrammarFilename = String(argtbl['grammar'])
    
    @ParamFilename = String(argtbl['param'])
    
    @AlgorithmName = String(argtbl['algo'] || DefaultAlgoName)
    
    @PrintInside = (argtbl['print-inside'])? true: false
    
    @PrintChart = (argtbl['print-chart'])? true: false
    
    @PrintParameter = (argtbl['print-param'])? true: false
    
    @PrintParser = (argtbl['print-parser'])? true: false
    
    arg = argtbl['epsilon']
    @Threshold = (arg)? Float(arg): DefaultThreshold
    @Threshold = nil if @Threshold == 0
    
    arg = argtbl['count']
    @Loopcount = (arg)? Integer(arg): DefaultLoopcount
    @Loopcount = nil if @Loopcount == 0
    
    raise(ArgumentError, "specify condition to terminate EM loop") if !@Threshold and !@Loopcount
  end
  
  # @f[^̓ǂݍ
  #----------------------------------------
  def read_pcfg ()
    @grammar = read_grammar(@GrammarFilename)
    raise(NoDataGivenException, "no grammar given") if !@grammar
    
    param = read_param(@ParamFilename)
    raise(NoDataGivenException, "no parameter given") if !param
    
    pcfg = Ext::PCFG.new(@grammar)
    param.each do |group|
      group.each do |pair|
        pcfg.store( pair.first, pair.last )
      end
    end
    
    return pcfg
  end
  
  # [eBeBIuWFNg̎s
  #----------------------------------------
  def call_body ( argtbl )
    scan_argument( argtbl )
    
    @total = 0
    
    pcfg = read_pcfg()
    
    case @AlgorithmName
    when 'fixed' then
      parser = Ext::Stolcke::FixedParser.new(pcfg)
    when 'flq' then
      parser = Ext::Stolcke::FLQParser.new(pcfg)
    when 'original' then
      parser = Ext::Stolcke::OriginalParser.new(pcfg)
      parser.pqueue_inside_test() if defined? parser.pqueue_inside_test
    else
      raise(InvalidAlgorithmException, "invalid algorithm name ``#{@AlgorithmName}''")
    end
    v_print(" + using algorithm = #{@AlgorithmName}\n")
    parser.display_parser(@stderr) if @PrintParser
    
    read_multi_corpus( @inputs ) {|corpus| process_corpus(parser, corpus) }
    
    # wKJn
    @start_time = Time.now()
    @start_ptime = get_process_time().ptime
    v_print(" + start EM learning\n")
    v_print("   - loop count = #{@Loopcount}\n") if @Loopcount
    v_print("   - threshold = #{@Threshold}\n") if @Threshold
    @history = [ ]
    while loop_continue?(parser) do
      parser.update()
      parser.display_chart(@stderr) if @PrintChart and defined?(parser.display_chart)
      parser.display_parameter(@stderr) if @PrintParameter and defined?(parser.display_parameter)
    end
    @end_time = Time.now()
    @end_ptime = get_process_time().ptime
    
    # wKʂ̏o
    result = parser.get_parameter
    result.display(@stdout)
    
    # Ȍo
    display_log()
    
    return self
  end
  attr_reader :total  # ǂݍ񂾕̑
  
  def loop_continue? ( parser )
    ptime = get_process_time().ptime # [vJn̎ԂĂ
    
    # m̌vZs
    curr_likelihood = parser.likelihood()
    parser.display_inside(@stderr) if @PrintInside and defined?(parser.display_inside)
    
    @history << EMHistory.new(@history.length + 1, curr_likelihood, Time.now, ptime)
    v_print {
      hist = @history.last
      sec_delta = (@history.length > 1)? @history[-1].sec - @history[-2].sec: 0
      sprintf("   #%5d: (%s) %e <%f sec>\n", hist.loopcount, hist.time.strftime("%m/%d %H:%M"), hist.likelihood, sec_delta)
    }
    
    if @Threshold then
      if @history.length > 1 then
        prev_likelihood = @history[-2].likelihood
        return false if curr_likelihood - prev_likelihood < @Threshold
      end
    end
    
    if @Loopcount then
      return false if @history.length >= @Loopcount
    end
    
    return true
  end
  
  def display_log ()
    timeformat = "%Y/%m/%d %H:%M:%S"
    
    start_time = @start_time.strftime(timeformat)
    end_time = @end_time.strftime(timeformat)
    diff_time = @end_time.tv_sec - @start_time.tv_sec
    
    hostname = ENV['HOST'] || ENV['HOSTNAME'] || `hostname`.strip
    
    user = ENV['USER'] || ENV['USERNAME'] || `whoami`.strip
    
    interval = (@end_ptime - @start_ptime) / @history.length
    
    Block.display('dEMlog', nil) do
      print "!user #{user}\n"
      print "!hostname #{hostname}\n"
      print "!time <#{start_time}> --- <#{end_time}> [#{diff_time} sec]\n"
      print "!interval #{interval}\n"
      print "!threshold #{@Threshold}\n" if @Threshold
      print "!loopcount #{@Loopcount}\n" if @Loopcount
      print "!history\n" \
            "\# loop\tprocess sec\tlikelihood\n"
      @history.each_with_index do |hist, i|
        proctime = (i > 0)? @history[i].sec - @history[i-1].sec: 0.0
        printf("%d\t%f\t%e\n", hist.loopcount, proctime, hist.likelihood)
      end
    end
  end
  
  # R[pXɑ΂鏈
  #----------------------------------------
  def process_corpus ( parser, corpus )
    print_progress = lambda do
      v_print {
        mem = Float(parser.memsize) / 1024
        if mem > 1024 then
          sprintf("   - \#%d (%.2f MB)\r", @total, mem / 1024)
        else
          sprintf("   - \#%d (%.1f KB)\r", @total, mem)
        end
      }
    end
    
    t = 0
    corpus.each do |sentence|
      @total += 1
      t = update_progress_info( t ) { print_progress.call() }
      begin
        # ̓ǂݍ
        itemset = read_sentence( sentence )
        break unless itemset
        
        # `[g̓o^
        parser.store(itemset)
      rescue
        print_warn(corpus.stream, "\##{@total}: #$!")
        t = 0
      end
    end # corpus.each
    print_progress.call() ; v_print("\n")
  end
  
  # <pSentence> ̓ǂݍŕKvȏW߂
  #----------------------------------------
  def read_sentence ( sentence )
    itemset = nil
    sentence.each(ItemSetBlock) do |cls, block|
      if cls then
        if cls == ItemSetBlock then
          raise(TooManyDataException, "Found too many <dItemSet>.") if itemset
          itemset = cls.new(block).read(@grammar)
        end
      else
        block.skip
      end
    end
    return itemset
  end
end


######################################################################
if $0 == __FILE__ then
### ̃t@Ĉ܂܎sꂽꍇ̏
print_error_message = lambda {|mes| $stderr << "ERROR: #{mes}\n" }

begin
  optspec = [ ]
  optspec << ['--grammar', '-g', ARG_RQ]
  optspec << ['--param', '-p', ARG_RQ]
  optspec << ['--algo', '-a', ARG_RQ]
  optspec << ['--epsilon', '-e', ARG_RQ]
  optspec << ['--count', '-c', ARG_RQ]
  optspec << ['--print-inside', ARG_NO] if $DEBUG
  optspec << ['--print-chart', ARG_NO] if $DEBUG
  optspec << ['--print-param', ARG_NO] if $DEBUG
  optspec << ['--print-parser', ARG_NO] if $DEBUG
  
  core = StolckeParserUtility.new()
  argtbl = { }
  inputs = core.fetch_options( ARGV, argtbl, optspec )
  if inputs then
    begin
      core.call( argtbl, inputs )
    rescue NoDataGivenException
      print_error_message.call( $! )
      core.print_usage( $stdout )
    end
  end
rescue
  print_error_message.call( $! )
end

###
end
######################################################################
